package de.fhkoeln.fb10.ala.puzzle.ui;

import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.border.TitledBorder;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.fhkoeln.fb10.ala.puzzle.domain.Position;
import de.fhkoeln.fb10.ala.puzzle.domain.Puzzle;
import de.fhkoeln.fb10.ala.puzzle.domain.PuzzleListener;
import de.fhkoeln.fb10.ala.puzzle.domain.PuzzleImpl;

public class PuzzlePane extends JPanel implements PuzzleListener {

	private static final long serialVersionUID = 8028154920227801951L;
	
	private Log log = LogFactory.getLog(PuzzlePane.class);
	
	private PuzzleImpl puzzleState;

	private Map<Integer, CellPane> cellPanes;
	
	private Integer padding;
	private Integer spacing;
	private String title;

	private Integer movementTime;
	private Integer frameRate = 20;	

	private BufferedImage image;
	
	//scale workCopy
	private BufferedImage imageWorkCopy = null;
	
	public PuzzlePane() {
	}

	public void initialize() {
		if(this.movementTime==null || movementTime<0)
			throw new IllegalArgumentException("Movementime cannot be null or smaller zero");
		
		if(this.puzzleState==null)
			throw new IllegalArgumentException("The PuzzlePane needs a Puzzle");
		if(this.title==null)
			throw new IllegalArgumentException("The PuzzlePane needs a Title");

		this.setLayout(null);
		this.addComponentListener(new PuzzlePaneListener());
		this.cellPanes = new HashMap<Integer, CellPane>();
		
		TitledBorder titledBorder = BorderFactory.createTitledBorder(null, this.title, TitledBorder.DEFAULT_JUSTIFICATION, TitledBorder.DEFAULT_POSITION, null, null);
		titledBorder.setTitleFont(new java.awt.Font("Arial", java.awt.Font.PLAIN, 10));
		this.setBorder(titledBorder);
		
		this.setVisible(true);
	}

	/**
	 * 
	 * @param key
	 * @return Direction this key has been moved to or -1 if could not move
	 */

	public void moveCell(CellPane c) {
		this.puzzleState.move(c.getKey());
	}
	
	public void moveCell(CellPane c, Integer dir) {

		Integer frameNumbers = (movementTime*frameRate)/1000;
		if(frameNumbers<1) {
			frameNumbers =1;
		}
		
		Integer startX = c.getX();
		Integer startY = c.getY();		
		
		Integer	destX = c.getX();
		Integer destY = c.getY();
		
		Integer curX = c.getX();
		Integer curY = c.getY();
		
		if(dir==Puzzle.DIR_LEFT) {
			destX = c.getX()-spacing/2-c.getWidth();
		} else if(dir==Puzzle.DIR_RIGHT) {
			destX = c.getX()+spacing/2+c.getWidth();
		} else if(dir==Puzzle.DIR_UP) {
			destY = c.getY()-c.getHeight()-spacing/2;
		} else if(dir==Puzzle.DIR_DOWN) {
			destY = c.getY()+c.getHeight()+spacing/2;
		} else	{
			return;
		}
		
		while(Math.abs(destX-curX)+Math.abs(destY-curY)>10) {
			Integer stepX = (destX-startX)/frameNumbers;
			Integer stepY = (destY-startY)/frameNumbers;
			
			curX += stepX;
			curY += stepY;
			
			c.setLocation(curX,curY);

			paint(getGraphics());
			
			try {
				Thread.sleep(movementTime/frameNumbers);
			} catch (InterruptedException e) {
				log.error(e);
			}
		}

		c.setLocation(destX,destY);
		paint(getGraphics());
	}

	private static GraphicsConfiguration getDefaultConfiguration() {
	    GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
	    GraphicsDevice gd = ge.getDefaultScreenDevice();
	    return gd.getDefaultConfiguration();
	}
	 
	private static BufferedImage copy(BufferedImage source, BufferedImage target) {
	    Graphics2D g2 = target.createGraphics();
	    g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
	    double scalex = (double) target.getWidth()/ source.getWidth();
	    double scaley = (double) target.getHeight()/ source.getHeight();
	    AffineTransform xform = AffineTransform.getScaleInstance(scalex, scaley);
	    g2.drawRenderedImage(source, xform);
	    g2.dispose();
	    return target;
	}
	 
	private static BufferedImage getScaledInstance(BufferedImage image, int width, int height, GraphicsConfiguration gc) {
	    if (gc == null)
	        gc = getDefaultConfiguration();
	    int transparency = image.getColorModel().getTransparency();
	    return copy(image, gc.createCompatibleImage(width, height, transparency));
	}
	
	public void drawPuzzle() {
		drawPuzzle(true);
	}
	
	public void drawPuzzle(boolean resize) {
		int[][] keys = puzzleState.getState();
		
		this.removeAll();
		this.cellPanes.clear();

		int width = (this.getSize().width-padding )/this.puzzleState.getSize();
		int height = (this.getSize().height-padding)/this.puzzleState.getSize();
		
		if(image != null) {
			if(resize)
				imageWorkCopy = getScaledInstance(image,this.getSize().width-padding,this.getSize().height-padding,getDefaultConfiguration());
			//log.info("New Image size: " + imageWorkCopy.getWidth() + " " + imageWorkCopy.getHeight());
		}
		
		for(int i=0;i<puzzleState.getSize();i++) {
			for(int j=0;j<puzzleState.getSize();j++) {
				if(keys[j][i]!=Puzzle.KEY_BLANK_STONE) {
					CellPane c = new CellPane(keys[j][i]);
					if(image != null && imageWorkCopy != null) {
						Position pos = new PuzzleImpl(puzzleState.getDefaultGoalState()).getPositionOf(keys[j][i]);
						c.setImage(imageWorkCopy.getSubimage(pos.getRow()*width,pos.getColumn()*height,width,height));
					}

					c.setPuzzlePane(this);
					c.setBounds(i*width+(padding/2), j*height+(padding/2), width-(spacing/2), height-(spacing/2));
					c.initialize();
					cellPanes.put(keys[j][i],c);
					
					this.add(c,null);
				}
			}
		}
		paint(this.getGraphics());
		updateUI();
	}

	public void puzzleUpdate(Puzzle puzzle, Object a) {
		List val = (List)a;
		moveCell(cellPanes.get(val.get(0)),(Integer)val.get(1));
	}
	
	public Integer getPadding() {
		return padding;
	}

	public void setPadding(Integer padding) {
		this.padding = padding;
	}

	public Integer getSpacing() {
		return spacing;
	}

	public void setSpacing(Integer spacing) {
		this.spacing = spacing;
	}
	
	public String getTitle() {
		return title;
	}

	public void setTitle(String title) {
		this.title = title;
	}

	public Map<Integer, CellPane> getCellPanes() {
		return cellPanes;
	}

	public void setCellPanes(Map<Integer, CellPane> cellPanes) {
		this.cellPanes = cellPanes;
	}

	public Integer getMovementTime() {
		return movementTime;
	}

	public void setMovementTime(Integer movementTime) {
		this.movementTime = movementTime;
	}

	public PuzzleImpl getPuzzleState() {
		return puzzleState;
	}

	public void setPuzzleState(PuzzleImpl puzzle) {
		this.puzzleState = puzzle;
	}
	public Integer getFrameRate() {
		return frameRate;
	}

	public void setFrameRate(Integer frameRate) {
		this.frameRate = frameRate;
	}

	public BufferedImage getImage() {
		return image;
	}

	public void setImage(BufferedImage image) {
		this.image = image;
	}	
}
